# cython: infer_types=True, profile=True, binding=True
from typing import Optional, Callable

import srsly
from thinc.api import Model, SequenceCategoricalCrossentropy, Config

from spacy.tokens.doc cimport Doc

from spacy.pipeline.tagger import Tagger
from spacy.language import Language
from spacy.errors import Errors
from spacy.scorer import Scorer
from spacy.tokens import Doc as PyDoc
from spacy.training import Example
from spacy.util import registry, get_words_and_spaces

from .scorers import tokenizer_senter_score


default_model_config = """
[model]
@architectures = "spacy.Tagger.v1"

[model.tok2vec]
@architectures = "spacy.Tok2Vec.v2"

[model.tok2vec.embed]
@architectures = "spacy.MultiHashEmbed.v2"
width = 128
attrs = ["ORTH","LOWER","IS_DIGIT","IS_ALPHA","IS_SPACE","IS_PUNCT"]
rows = [1000,500,50,50,50,50]
include_static_vectors = false

[model.tok2vec.encode]
@architectures = "spacy.MaxoutWindowEncoder.v2"
width = 128
depth = 4
window_size = 4
maxout_pieces = 2
"""
DEFAULT_CHAR_TAGGER_TOKENIZER_MODEL = Config().from_str(default_model_config)["model"]


@Language.factory(
    "experimental_char_tagger_tokenizer",
    assigns=[],
    default_config={"model": DEFAULT_CHAR_TAGGER_TOKENIZER_MODEL, "annotate_sents": True, "scorer": {"@scorers": "spacy-experimental.tokenizer_senter_scorer.v1"}},
    default_score_weights={"token_f": 0.5, "token_p": 0.0, "token_r": 0.0, "token_acc": None, "sents_f": 1.0, "sents_p": 0.0, "sents_r": 0.0},
    retokenizes=True,
)
def make_char_tagger_tokenizer(nlp: Language, name: str, model: Model, annotate_sents: bool, scorer: Optional[Callable]):
    return CharTaggerTokenizer(nlp.vocab, model, name, annotate_sents=annotate_sents, scorer=scorer)


class CharTaggerTokenizer(Tagger):
    def __init__(
        self,
        vocab,
        model,
        name="char_tagger_tokenizer",
        *,
        annotate_sents=True,
        scorer=tokenizer_senter_score,
    ):
        super().__init__(vocab, model, name, overwrite=True, scorer=scorer)
        self.cfg["annotate_sents"] = annotate_sents

    def set_annotations(self, docs, batch_tag_ids):
        if isinstance(docs, Doc):
            docs = [docs]
        cdef Doc doc
        cdef int idx, token_start
        labels = self.labels
        sent_label_id = -1
        if "S" in labels:
            sent_label_id = labels.index("S")
        token_label_id = labels.index("T")
        for i, doc in enumerate(docs):
            orig_text = doc.text
            doc_tag_ids = batch_tag_ids[i]
            if hasattr(doc_tag_ids, "get"):
                doc_tag_ids = doc_tag_ids.get()
            sent_starts = set()
            if self.cfg["annotate_sents"]:
                sent_starts = {j for j, tag_id in enumerate(doc_tag_ids) if tag_id == sent_label_id}
                # first character always starts a sentence
                sent_starts.add(0)
            token_start = 0
            with doc.retokenize() as retokenizer:
                for j, tag_id in enumerate(doc_tag_ids[1:], start=1):
                    if tag_id == token_label_id or tag_id == sent_label_id:
                        retokenizer.merge(doc[token_start:j])
                        token_start = j
            # try to merge any remaining trailing SPACY spaces
            with doc.retokenize() as retokenizer:
                j = 0
                while j < doc.length - 1:
                    if doc[j + 1].text == " " and \
                            not doc[j].is_space and \
                            not doc[j].text.endswith(" "):
                        retokenizer.merge(doc[j:j+2])
                        j += 1
                    j += 1
            # reset all the tag values
            j = 0
            while j < doc.length:
                text = doc[j].text
                if len(text) > 1 and text.endswith(" "):
                    lex = doc.vocab.get(doc.vocab.mem, text[:-1])
                    doc.c[j].lex = lex
                    doc.c[j].spacy = True
                if self.cfg["annotate_sents"]:
                    if doc.c[j].idx in sent_starts:
                        doc.c[j].sent_start = 1
                    else:
                        doc.c[j].sent_start = -1
                doc.c[j].tag = 0
                j += 1
            assert doc.text == orig_text

    def update(self, examples, *, **kwargs):
        examples = [Example(example.predicted, self._convert_to_char_doc(example.reference)) for example in examples]
        super().update(examples, **kwargs)

    def initialize(self, get_examples, *, **kwargs):
        examples = [Example(example.predicted, self._convert_to_char_doc(example.reference)) for example in get_examples()]
        super().initialize(lambda: examples, **kwargs)

    def _convert_to_char_doc(self, doc: Doc):
        """Convert a doc to character tokenization with TAG values for the
        tokenization and sentence segmentation (if enabled):
        - S: character starts a sentence
        - T: character starts a token (overridden by S if enabled)
        - I: character is within a token
        - O: character is not part of a token
        """
        words = list(doc.text)
        char_doc = PyDoc(doc.vocab, words=words, spaces=[False] * len(words))
        if not doc.has_unknown_spaces:
            for char_token in char_doc:
                char_token.tag_ = "O"
            for token in doc:
                char_doc[token.idx].tag_ = "T"
                if self.cfg["annotate_sents"]:
                    if token.is_sent_start:
                        char_doc[token.idx].tag_ = "S"
                for idx in range(token.idx + 1, token.idx + len(token.text)):
                    char_doc[idx].tag_ = "I"
        return char_doc



from typing import Callable
from spacy.language import Language
from spacy.tokens import Doc
from spacy.util import DummyTokenizer
from spacy.vocab import Vocab
from spacy import util


def char_pretokenizer_v1() -> Callable[[Language], "CharPretokenizer"]:
    """Function to create a character tokenizer.
    """
    def char_pretokenizer_factory(nlp: Language) -> "CharPretokenizer":
        return CharPretokenizer(nlp.vocab)

    return char_pretokenizer_factory


class CharPretokenizer(DummyTokenizer):
    def __init__(self, vocab: Vocab):
        self.vocab = vocab

    def __call__(self, string: str):
        # split into individual characters
        words = list(string)
        return Doc(self.vocab, words=words, spaces=[False] * len(words))



# cython: infer_types=True, profile=True, binding=True
from typing import Optional, Callable

import srsly
from thinc.api import Model, SequenceCategoricalCrossentropy, Config

from spacy.tokens.doc cimport Doc

from spacy.pipeline.ner import EntityRecognizer
from spacy.language import Language
from spacy.scorer import Scorer
from spacy.tokens import Doc as PyDoc, Span
from spacy.training import Example
from spacy.util import registry

from .scorers import tokenizer_score


default_model_config = """
[model]
@architectures = "spacy.TransitionBasedParser.v2"
state_type = "ner"
extra_state_tokens = false
hidden_width = 64
maxout_pieces = 2
use_upper = true
nO = null

[model.tok2vec]
@architectures = "spacy.Tok2Vec.v2"

[model.tok2vec.embed]
@architectures = "spacy.MultiHashEmbed.v2"
width = 128
attrs = ["ORTH","LOWER","IS_DIGIT","IS_ALPHA","IS_SPACE","IS_PUNCT"]
rows = [1000,500,50,50,50,50]
include_static_vectors = false

[model.tok2vec.encode]
@architectures = "spacy.MaxoutWindowEncoder.v2"
width = 128
depth = 4
window_size = 4
maxout_pieces = 2
"""
DEFAULT_CHAR_NER_TOKENIZER_MODEL = Config().from_str(default_model_config)["model"]


@Language.factory(
    "experimental_char_ner_tokenizer",
    assigns=["token.is_sent_start"],
    default_config={"model": DEFAULT_CHAR_NER_TOKENIZER_MODEL, "scorer": {"@scorers": "spacy-experimental.tokenizer_scorer.v1"}},
    default_score_weights={"token_f": 0.5, "token_p": 0.0, "token_r": 0.0, "token_acc": None},
    retokenizes=True,
)
def make_char_ner_tokenizer(nlp: Language, name: str, model: Model, scorer: Optional[Callable]):
    return CharNERTokenizer(nlp.vocab, model, name, scorer=scorer)


class CharNERTokenizer(EntityRecognizer):
    def __init__(
        self,
        vocab,
        model,
        name="exp_char_ner_tokenizer",
        *,
        scorer=tokenizer_score,
    ):
        super().__init__(vocab, model, name, None, scorer=scorer)

    def set_annotations(self, docs, scores):
        super().set_annotations(docs, scores)
        cdef Doc doc
        cdef int j, k, end
        for doc in docs:
            orig_text = doc.text
            # merge all tokens marked as ents with trailing space if present
            with doc.retokenize() as retokenizer:
                for ent in doc.ents:
                    end = ent.end
                    if ent.end < len(doc) and doc[ent.end].text == " ":
                        end += 1
                    retokenizer.merge(doc[ent.start:end])
            # try to merge any remaining trailing SPACY spaces
            with doc.retokenize() as retokenizer:
                j = 0
                while j < doc.length - 1:
                    if doc[j + 1].text == " " and \
                            not doc[j].is_space and \
                            not doc[j].text.endswith(" "):
                        retokenizer.merge(doc[j:j+2])
                        j += 1
                    j += 1
            doc.set_ents([], default="missing")
            j = 0
            while j < doc.length:
                text = doc[j].text
                if len(text) > 1 and text.endswith(" ") and not text.isspace():
                    lex = doc.vocab.get(doc.vocab.mem, text[:-1])
                    doc.c[j].lex = lex
                    doc.c[j].spacy = True
                j += 1
            assert doc.text == orig_text

    def update(self, examples, *, **kwargs):
        examples = [Example(example.predicted, self._convert_to_char_doc(example.reference)) for example in examples]
        super().update(examples, **kwargs)

    def initialize(self, get_examples, *, **kwargs):
        examples = [Example(example.predicted, self._convert_to_char_doc(example.reference)) for example in get_examples()]
        super().initialize(lambda: examples, **kwargs)

    def _convert_to_char_doc(self, doc: Doc):
        words = list(doc.text)
        char_doc = PyDoc(doc.vocab, words=words, spaces=[False] * len(words))
        if not doc.has_unknown_spaces:
            char_doc.ents = [
                Span(char_doc, token.idx, token.idx+len(token.text), label="TOKEN")
                for token in doc
            ]
        return char_doc



import pytest
from spacy import util
from spacy.lang.en import English
from spacy.language import Language
from spacy.training import Example


TRAIN_DATA = [
    ("She likes green eggs", {"lemmas": ["she", "like", "green", "egg"]}),
    ("Eat blue ham", {"lemmas": ["eat", "blue", "ham"]}),
]

PARTIAL_DATA = [
    # partial annotation
    ("She likes green eggs", {"lemmas": ["", "like", "green", ""]}),
    # misaligned partial annotation
    (
        "He hates green eggs",
        {
            "words": ["He", "hat", "es", "green", "eggs"],
            "lemmas": ["", "hat", "e", "green", ""],
        },
    ),
]


def test_initialize_examples():
    nlp = Language()
    lemmatizer = nlp.add_pipe("experimental_edit_tree_lemmatizer")
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
    # you shouldn't really call this more than once, but for testing it should be fine
    nlp.initialize(get_examples=lambda: train_examples)
    with pytest.raises(TypeError):
        nlp.initialize(get_examples=lambda: None)
    with pytest.raises(TypeError):
        nlp.initialize(get_examples=lambda: train_examples[0])
    with pytest.raises(TypeError):
        nlp.initialize(get_examples=lambda: [])
    with pytest.raises(TypeError):
        nlp.initialize(get_examples=train_examples)


def test_no_data():
    # Test that the lemmatizer provides a nice error when there's no tagging data / labels
    TEXTCAT_DATA = [
        ("I'm so happy.", {"cats": {"POSITIVE": 1.0, "NEGATIVE": 0.0}}),
        ("I'm so angry", {"cats": {"POSITIVE": 0.0, "NEGATIVE": 1.0}}),
    ]
    nlp = English()
    nlp.add_pipe("experimental_edit_tree_lemmatizer")
    nlp.add_pipe("textcat")

    train_examples = []
    for t in TEXTCAT_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))

    with pytest.raises(ValueError):
        nlp.initialize(get_examples=lambda: train_examples)


def test_incomplete_data():
    # Test that the lemmatizer works with incomplete information
    nlp = English()
    lemmatizer = nlp.add_pipe("experimental_edit_tree_lemmatizer")
    lemmatizer.min_tree_freq = 1
    train_examples = []
    for t in PARTIAL_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))
    optimizer = nlp.initialize(get_examples=lambda: train_examples)
    for i in range(50):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)
    assert losses["experimental_edit_tree_lemmatizer"] < 0.00001

    # test the trained model
    test_text = "She likes blue eggs"
    doc = nlp(test_text)
    assert doc[1].lemma_ == "like"
    assert doc[2].lemma_ == "blue"


def test_overfitting_IO():
    nlp = English()
    lemmatizer = nlp.add_pipe("experimental_edit_tree_lemmatizer")
    lemmatizer.min_tree_freq = 1
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))

    optimizer = nlp.initialize(get_examples=lambda: train_examples)

    for i in range(50):
        losses = {}
        nlp.update(train_examples, sgd=optimizer, losses=losses)
    assert losses["experimental_edit_tree_lemmatizer"] < 0.00001

    test_text = "She likes blue eggs"
    doc = nlp(test_text)
    assert doc[0].lemma_ == "she"
    assert doc[1].lemma_ == "like"
    assert doc[2].lemma_ == "blue"
    assert doc[3].lemma_ == "egg"

    # Check model after a {to,from}_disk roundtrip
    with util.make_tempdir() as tmp_dir:
        nlp.to_disk(tmp_dir)
        nlp2 = util.load_model_from_path(tmp_dir)
        doc2 = nlp2(test_text)
        assert doc2[0].lemma_ == "she"
        assert doc2[1].lemma_ == "like"
        assert doc2[2].lemma_ == "blue"
        assert doc2[3].lemma_ == "egg"

    # Check model after a {to,from}_bytes roundtrip
    nlp_bytes = nlp.to_bytes()
    nlp3 = English()
    nlp3.add_pipe("experimental_edit_tree_lemmatizer")
    nlp3.from_bytes(nlp_bytes)
    doc3 = nlp3(test_text)
    assert doc3[0].lemma_ == "she"
    assert doc3[1].lemma_ == "like"
    assert doc3[2].lemma_ == "blue"
    assert doc3[3].lemma_ == "egg"


def test_lemmatizer_requires_labels():
    nlp = English()
    nlp.add_pipe("experimental_edit_tree_lemmatizer")
    with pytest.raises(ValueError):
        nlp.initialize()


def test_lemmatizer_label_data():
    nlp = English()
    lemmatizer = nlp.add_pipe("experimental_edit_tree_lemmatizer")
    lemmatizer.min_tree_freq = 1
    train_examples = []
    for t in TRAIN_DATA:
        train_examples.append(Example.from_dict(nlp.make_doc(t[0]), t[1]))

    nlp.initialize(get_examples=lambda: train_examples)

    nlp2 = English()
    lemmatizer2 = nlp2.add_pipe("experimental_edit_tree_lemmatizer")
    lemmatizer2.initialize(
        get_examples=lambda: train_examples, labels=lemmatizer.label_data
    )

    # Verify that the labels and trees are the same.
    assert lemmatizer.labels == lemmatizer2.labels
    assert lemmatizer.trees.to_bytes() == lemmatizer2.trees.to_bytes()



from spacy.scorer import Scorer
from spacy.util import registry


def tokenizer_score(examples, **kwargs):
    return Scorer.score_tokenization(examples)


def make_tokenizer_scorer():
    return tokenizer_score


def tokenizer_senter_score(examples, **kwargs):
    def has_sents(doc):
        return doc.has_annotation("SENT_START")

    results = Scorer.score_tokenization(examples)
    results.update(Scorer.score_spans(examples, "sents", has_annotation=has_sents, **kwargs))
    del results["sents_per_type"]
    return results


def make_tokenizer_senter_scorer():
    return tokenizer_senter_score



from hypothesis import given
import hypothesis.strategies as st
from spacy.strings import StringStore
from spacy.util import make_tempdir

from spacy_experimental.edit_tree_lemmatizer.edit_trees import EditTrees


def test_dutch():
    strings = StringStore()
    trees = EditTrees(strings)
    tree = trees.add("deelt", "delen")
    assert trees.tree_to_str(tree) == "(m 0 3 () (m 0 2 (s '' 'l') (s 'lt' 'n')))"

    tree = trees.add("gedeeld", "delen")
    assert (
        trees.tree_to_str(tree) == "(m 2 3 (s 'ge' '') (m 0 2 (s '' 'l') (s 'ld' 'n')))"
    )


def test_from_to_bytes():
    strings = StringStore()
    trees = EditTrees(strings)
    trees.add("deelt", "delen")
    trees.add("gedeeld", "delen")

    b = trees.to_bytes()

    trees2 = EditTrees(strings)
    trees2.from_bytes(b)

    # Verify that the nodes did not change.
    assert len(trees) == len(trees2)
    for i in range(len(trees)):
        assert trees.tree_to_str(i) == trees2.tree_to_str(i)

    # Reinserting the same trees should not add new nodes.
    trees2.add("deelt", "delen")
    trees2.add("gedeeld", "delen")
    assert len(trees) == len(trees2)


def test_from_to_disk():
    strings = StringStore()
    trees = EditTrees(strings)
    trees.add("deelt", "delen")
    trees.add("gedeeld", "delen")

    trees2 = EditTrees(strings)
    with make_tempdir() as temp_dir:
        trees_file = temp_dir / "edit_trees.bin"
        trees.to_disk(trees_file)
        trees2 = trees2.from_disk(trees_file)

    # Verify that the nodes did not change.
    assert len(trees) == len(trees2)
    for i in range(len(trees)):
        assert trees.tree_to_str(i) == trees2.tree_to_str(i)

    # Reinserting the same trees should not add new nodes.
    trees2.add("deelt", "delen")
    trees2.add("gedeeld", "delen")
    assert len(trees) == len(trees2)


@given(st.text(), st.text())
def test_roundtrip(form, lemma):
    strings = StringStore()
    trees = EditTrees(strings)
    tree = trees.add(form, lemma)
    assert trees.apply(tree, form) == lemma


@given(st.text(alphabet="ab"), st.text(alphabet="ab"))
def test_roundtrip_small_alphabet(form, lemma):
    # Test with small alphabets to have more overlap.
    strings = StringStore()
    trees = EditTrees(strings)
    tree = trees.add(form, lemma)
    assert trees.apply(tree, form) == lemma


def test_unapplicable_trees():
    strings = StringStore()
    trees = EditTrees(strings)
    tree3 = trees.add("deelt", "delen")

    # Replacement fails.
    assert trees.apply(tree3, "deeld") == None

    # Suffix + prefix are too large.
    assert trees.apply(tree3, "de") == None


def test_empty_strings():
    strings = StringStore()
    trees = EditTrees(strings)
    no_change = trees.add("xyz", "xyz")
    empty = trees.add("", "")
    assert no_change == empty



from setuptools import setup, Extension, find_packages
from distutils.command.build_ext import build_ext
from Cython.Build import cythonize
from Cython.Compiler import Options
import numpy


# Preserve `__doc__` on functions and classes
# http://docs.cython.org/en/latest/src/userguide/source_files_and_compilation.html#compiler-options
Options.docstrings = True

COMPILE_OPTIONS = {
    "msvc": ["/Ox", "/EHsc"],
    "mingw32": ["-O2", "-Wno-strict-prototypes", "-Wno-unused-function"],
    "other": ["-O2", "-Wno-strict-prototypes", "-Wno-unused-function"],
}
LINK_OPTIONS = {"msvc": ["-std=c++11"], "mingw32": ["-std=c++11"], "other": []}
COMPILER_DIRECTIVES = {
    "language_level": -3,
    "embedsignature": True,
    "annotation_typing": False,
}


# By subclassing build_extensions we have the actual compiler that will be used which is really known only after finalize_options
# http://stackoverflow.com/questions/724664/python-distutils-how-to-get-a-compiler-that-is-going-to-be-used
class build_ext_options:
    def build_options(self):
        for e in self.extensions:
            e.extra_compile_args += COMPILE_OPTIONS.get(
                self.compiler.compiler_type, COMPILE_OPTIONS["other"]
            )
        for e in self.extensions:
            e.extra_link_args += LINK_OPTIONS.get(
                self.compiler.compiler_type, LINK_OPTIONS["other"]
            )


class build_ext_subclass(build_ext, build_ext_options):
    def build_extensions(self):
        build_ext_options.build_options(self)
        build_ext.build_extensions(self)


def setup_package():
    ext_modules = [
        Extension(
            "spacy_experimental.edit_tree_lemmatizer.edit_trees",
            ["spacy_experimental/edit_tree_lemmatizer/edit_trees.pyx"],
            language="c++",
            include_dirs=[],
            extra_compile_args=["-std=c++11"],
        ),
                Extension(
            "spacy_experimental.char_tokenizer.char_tagger_tokenizer",
            ["spacy_experimental/char_tokenizer/char_tagger_tokenizer.pyx"],
            language="c++",
            include_dirs=[numpy.get_include()],
            extra_compile_args=["-std=c++11"],
        ),
        Extension(
            "spacy_experimental.char_tokenizer.char_ner_tokenizer",
            ["spacy_experimental/char_tokenizer/char_ner_tokenizer.pyx"],
            language="c++",
            include_dirs=[numpy.get_include()],
            extra_compile_args=["-std=c++11"],
        ),
    ]

    ext_modules = cythonize(ext_modules, compiler_directives=COMPILER_DIRECTIVES)

    setup(
        name="spacy-experimental",
        packages=find_packages(),
        ext_modules=ext_modules,
        cmdclass={"build_ext": build_ext_subclass},
        package_data={"": ["*.pyx", "*.pxd", "*.pxi"]},
    )


if __name__ == "__main__":
    setup_package()



import pytest
import spacy
from spacy import util
from spacy.training import Example


@pytest.mark.parametrize("pipe_name", ["experimental_char_tagger_tokenizer", "experimental_char_ner_tokenizer"])
def test_char_tokenizer_overfitting(pipe_name):
    # learn the default English tokenization
    texts = [
        "This is a sentence.",
        "Here is a short, boring sentence.",
        "Here is another!",
    ]
    nlp = spacy.blank("en")
    train_docs = [nlp.make_doc(text) for text in texts]

    nlp = spacy.blank(
        "en",
        vocab=nlp.vocab,
        config={
            "nlp": {
                "tokenizer": {"@tokenizers": "spacy-experimental.char_pretokenizer.v1"}
            }
        },
    )
    nlp.add_pipe(pipe_name)
    train_examples = [Example(nlp.make_doc(doc.text), doc) for doc in train_docs]
    optimizer = nlp.initialize(get_examples=lambda: train_examples)

    for i in range(50):
        nlp.update(train_examples, sgd=optimizer)

    for train_doc, test_doc in zip(train_docs, nlp.pipe(texts)):
        assert train_doc.text == test_doc.text
        assert [t.text for t in train_doc] == [t.text for t in test_doc]
